@@ -9,7 +9,8 @@ Azure IoTHub Device SDK for Python.
 
 import logging
 import asyncio
-from azure.iot.device.common import async_adapter, asyncio_compat
+import deprecation
+from azure.iot.device.common import async_adapter
 from azure.iot.device.iothub.abstract_clients import (
     AbstractIoTHubClient,
     AbstractIoTHubDeviceClient,
@@ -21,7 +22,7 @@ from azure.iot.device.iothub.pipeline import exceptions as pipeline_exceptions
 from azure.iot.device import exceptions
 from azure.iot.device.iothub.inbox_manager import InboxManager
 from .async_inbox import AsyncClientInbox
-from . import async_handler_manager
+from . import async_handler_manager, loop_management
 from azure.iot.device import constant as device_constant
 
 logger = logging.getLogger(__name__)
@@ -80,16 +81,6 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         self._mqtt_pipeline.on_method_request_received = self._inbox_manager.route_method_request
         self._mqtt_pipeline.on_twin_patch_received = self._inbox_manager.route_twin_patch
 
-    def _on_connected(self):
-        """Helper handler that is called upon an iothub pipeline connect"""
-        logger.info("Connection State - Connected")
-
-    def _on_disconnected(self):
-        """Helper handler that is called upon an iothub pipeline disconnect"""
-        logger.info("Connection State - Disconnected")
-        self._inbox_manager.clear_all_method_requests()
-        logger.info("Cleared all pending method requests due to disconnect")
-
     async def _enable_feature(self, feature_name):
         """Enable an Azure IoT Hub feature
 
@@ -128,7 +119,7 @@ class GenericIoTHubClient(AbstractIoTHubClient):
             logger.info("Successfully disabled feature: {}".format(feature_name))
         else:
             # This branch shouldn't be reached, but in case it is, log it
-            logger.info("Feature ({}) already enabled - skipping".format(feature_name))
+            logger.info("Feature ({}) already disabled - skipping".format(feature_name))
 
     async def connect(self):
         """Connects the client to an Azure IoT Hub or Azure IoT Edge Hub instance.
@@ -157,15 +148,43 @@ class GenericIoTHubClient(AbstractIoTHubClient):
     async def disconnect(self):
         """Disconnect the client from the Azure IoT Hub or Azure IoT Edge Hub instance.
 
+        It is recommended that you make sure to call this coroutine when you are completely done
+        with the your client instance.
+
         :raises: :class:`azure.iot.device.exceptions.ClientError` if there is an unexpected failure
             during execution.
         """
         logger.info("Disconnecting from Hub...")
-        disconnect_async = async_adapter.emulate_async(self._mqtt_pipeline.disconnect)
 
+        logger.debug("Executing initial disconnect")
+        disconnect_async = async_adapter.emulate_async(self._mqtt_pipeline.disconnect)
         callback = async_adapter.AwaitableCallback()
         await disconnect_async(callback=callback)
         await handle_result(callback)
+        logger.debug("Successfully executed initial disconnect")
+
+        # Note that in the process of stopping the handlers and resolving pending calls
+        # a user-supplied handler may cause a reconnection to occur
+        logger.debug("Stopping handlers...")
+        self._handler_manager.stop()
+        logger.debug("Successfully stopped handlers")
+
+        # Disconnect again to ensure disconnection has ocurred due to the issue mentioned above
+        logger.debug("Executing secondary disconnect...")
+        disconnect_async = async_adapter.emulate_async(self._mqtt_pipeline.disconnect)
+        callback = async_adapter.AwaitableCallback()
+        await disconnect_async(callback=callback)
+        await handle_result(callback)
+        logger.debug("Successfully executed secondary disconnect")
+
+        # It's also possible that in the (very short) time between stopping the handlers and
+        # the second disconnect, additional items were received (e.g. C2D Message)
+        # Currently, this isn't really possible to accurately check due to a race condition.
+        # It has always been true of this client, even before handlers.
+        # TODO: fix the race condition
+        # However, even if the race condition is addressed, that will only allow us to log that
+        # messages were lost. To actually fix the problem, IoTHub needs to support MQTT5 so that
+        # we can unsubscribe from receiving data.
 
         logger.info("Successfully disconnected from Hub")
 
@@ -204,6 +223,11 @@ class GenericIoTHubClient(AbstractIoTHubClient):
 
         logger.info("Successfully sent message to Hub")
 
+    @deprecation.deprecated(
+        deprecated_in="2.3.0",
+        current_version=device_constant.VERSION,
+        details="We recommend that you use the .on_method_request_received property to set a handler instead",
+    )
     async def receive_method_request(self, method_name=None):
         """Receive a method request via the Azure IoT Hub or Azure IoT Edge Hub.
 
@@ -214,9 +238,9 @@ class GenericIoTHubClient(AbstractIoTHubClient):
             a different call to receive_method will be received.
 
         :returns: MethodRequest object representing the received method request.
-        :rtype: `azure.iot.device.MethodRequest`
+        :rtype: :class:`azure.iot.device.MethodRequest`
         """
-        self._validate_receive_api_invoke()
+        self._check_receive_mode_is_api()
 
         if not self._mqtt_pipeline.feature_enabled[constant.METHODS]:
             await self._enable_feature(constant.METHODS)
@@ -322,6 +346,11 @@ class GenericIoTHubClient(AbstractIoTHubClient):
 
         logger.info("Successfully sent twin patch")
 
+    @deprecation.deprecated(
+        deprecated_in="2.3.0",
+        current_version=device_constant.VERSION,
+        details="We recommend that you use the .on_twin_desired_properties_patch_received property to set a handler instead",
+    )
     async def receive_twin_desired_properties_patch(self):
         """
         Receive a desired property patch via the Azure IoT Hub or Azure IoT Edge Hub.
@@ -331,7 +360,7 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         :returns: Twin Desired Properties patch as a JSON dict
         :rtype: dict
         """
-        self._validate_receive_api_invoke()
+        self._check_receive_mode_is_api()
 
         if not self._mqtt_pipeline.feature_enabled[constant.TWIN_PATCHES]:
             await self._enable_feature(constant.TWIN_PATCHES)
@@ -342,6 +371,99 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         logger.info("twin patch received")
         return patch
 
+    def _generic_handler_setter(self, handler_name, feature_name, new_handler):
+        self._check_receive_mode_is_handler()
+        # Set the handler on the handler manager
+        setattr(self._handler_manager, handler_name, new_handler)
+
+        # Enable the feature if necessary
+        if new_handler is not None and not self._mqtt_pipeline.feature_enabled[feature_name]:
+            # We have to call this on a loop running on a different thread in order to ensure
+            # the setter can be called both within a coroutine (with a running event loop) and
+            # outside of a coroutine (where no event loop is currently running)
+            loop = loop_management.get_client_internal_loop()
+            fut = asyncio.run_coroutine_threadsafe(self._enable_feature(feature_name), loop=loop)
+            fut.result()
+
+        # Disable the feature if necessary
+        elif new_handler is None and self._mqtt_pipeline.feature_enabled[feature_name]:
+            # We have to call this on a loop running on a different thread in order to ensure
+            # the setter can be called both within a coroutine (with a running event loop) and
+            # outside of a coroutine (where no event loop is currently running)
+            loop = loop_management.get_client_internal_loop()
+            fut = asyncio.run_coroutine_threadsafe(self._disable_feature(feature_name), loop=loop)
+            fut.result()
+
+    @property
+    def on_twin_desired_properties_patch_received(self):
+        """The handler function or coroutine that will be called when a twin desired properties
+        patch is received.
+
+        The function or coroutine definition should take one positional argument (the twin patch
+        in the form of a JSON dictionary object)"""
+        return self._handler_manager.on_twin_desired_properties_patch_received
+
+    @on_twin_desired_properties_patch_received.setter
+    def on_twin_desired_properties_patch_received(self, value):
+        self._generic_handler_setter(
+            "on_twin_desired_properties_patch_received", constant.TWIN_PATCHES, value
+        )
+
+    @property
+    def on_method_request_received(self):
+        """The handler function or coroutine that will be called when a method request is received.
+
+        The function or coroutine definition should take one positional argument (the
+        :class:`azure.iot.device.MethodRequest` object)"""
+        return self._handler_manager.on_method_request_received
+
+    @on_method_request_received.setter
+    def on_method_request_received(self, value):
+        self._generic_handler_setter("on_method_request_received", constant.METHODS, value)
+
+
+class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
+    """An asynchronous device client that connects to an Azure IoT Hub instance.
+
+    Intended for usage with Python 3.5.3+
+    """
+
+    def __init__(self, mqtt_pipeline, http_pipeline):
+        """Initializer for a IoTHubDeviceClient.
+
+        This initializer should not be called directly.
+        Instead, use one of the 'create_from_' classmethods to instantiate
+
+        :param mqtt_pipeline: The pipeline used to connect to the IoTHub endpoint.
+        :type mqtt_pipeline: :class:`azure.iot.device.iothub.pipeline.MQTTPipeline`
+        """
+        super().__init__(mqtt_pipeline=mqtt_pipeline, http_pipeline=http_pipeline)
+        self._mqtt_pipeline.on_c2d_message_received = self._inbox_manager.route_c2d_message
+
+    @deprecation.deprecated(
+        deprecated_in="2.3.0",
+        current_version=device_constant.VERSION,
+        details="We recommend that you use the .on_message_received property to set a handler instead",
+    )
+    async def receive_message(self):
+        """Receive a message that has been sent from the Azure IoT Hub.
+
+        If no message is yet available, will wait until an item is available.
+
+        :returns: Message that was sent from the Azure IoT Hub.
+        :rtype: :class:`azure.iot.device.Message`
+        """
+        self._check_receive_mode_is_api()
+
+        if not self._mqtt_pipeline.feature_enabled[constant.C2D_MSG]:
+            await self._enable_feature(constant.C2D_MSG)
+        c2d_inbox = self._inbox_manager.get_c2d_message_inbox()
+
+        logger.info("Waiting for message from Hub...")
+        message = await c2d_inbox.get()
+        logger.info("Message received")
+        return message
+
     async def get_storage_info_for_blob(self, blob_name):
         """Sends a POST request over HTTP to an IoTHub endpoint that will return information for uploading via the Azure Storage Account linked to the IoTHub your device is connected to.
 
@@ -384,63 +506,17 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         await handle_result(callback)
         logger.info("Successfully notified blob upload status")
 
-    # @property
-    # def on_method_request_received(self):
-    #     return self._on_method_request_received
+    @property
+    def on_message_received(self):
+        """The handler function or coroutine that will be called when a message is received.
 
-    # @on_method_request_received.setter
-    # def on_method_request_received(self, value):
-    #     self._validate_receive_handler_setter()
-    #     # Enable the feature if necessary
-    #     if value is not None and not self._mqtt_pipeline.feature_enabled[constant.METHODS]:
-    #         loop = asyncio.get_event_loop()
-    #         asyncio.ensure_future(self._enable_feature(constant.METHODS), loop=loop)
+        The function or coroutine definition should take one positional argument (the
+        :class:`azure.iot.device.Message` object)"""
+        return self._handler_manager.on_message_received
 
-    #     # Disable the feature if necessary
-    #     elif value is None and self._mqtt_pipeline.feature_enabled[constant.METHODS]:
-    #         loop = asyncio.get_event_loop()
-    #         asyncio.ensure_future(self._disable_feature(constant.METHODS), loop=loop)
-
-    #     # Set the handler on the handler manager
-    #     self._handler_manager.on_method_request_received = value
-
-
-class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
-    """An asynchronous device client that connects to an Azure IoT Hub instance.
-
-    Intended for usage with Python 3.5.3+
-    """
-
-    def __init__(self, mqtt_pipeline, http_pipeline):
-        """Initializer for a IoTHubDeviceClient.
-
-        This initializer should not be called directly.
-        Instead, use one of the 'create_from_' classmethods to instantiate
-
-        :param mqtt_pipeline: The pipeline used to connect to the IoTHub endpoint.
-        :type mqtt_pipeline: :class:`azure.iot.device.iothub.pipeline.MQTTPipeline`
-        """
-        super().__init__(mqtt_pipeline=mqtt_pipeline, http_pipeline=http_pipeline)
-        self._mqtt_pipeline.on_c2d_message_received = self._inbox_manager.route_c2d_message
-
-    async def receive_message(self):
-        """Receive a message that has been sent from the Azure IoT Hub.
-
-        If no message is yet available, will wait until an item is available.
-
-        :returns: Message that was sent from the Azure IoT Hub.
-        :rtype: :class:`azure.iot.device.Message`
-        """
-        self._validate_receive_api_invoke()
-
-        if not self._mqtt_pipeline.feature_enabled[constant.C2D_MSG]:
-            await self._enable_feature(constant.C2D_MSG)
-        c2d_inbox = self._inbox_manager.get_c2d_message_inbox()
-
-        logger.info("Waiting for message from Hub...")
-        message = await c2d_inbox.get()
-        logger.info("Message received")
-        return message
+    @on_message_received.setter
+    def on_message_received(self, value):
+        self._generic_handler_setter("on_message_received", constant.C2D_MSG, value)
 
 
 class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
@@ -503,6 +579,11 @@ class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
 
         logger.info("Successfully sent message to output: " + output_name)
 
+    @deprecation.deprecated(
+        deprecated_in="2.3.0",
+        current_version=device_constant.VERSION,
+        details="We recommend that you use the .on_message_received property to set a handler instead",
+    )
     async def receive_message_on_input(self, input_name):
         """Receive an input message that has been sent from another Module to a specific input.
 
@@ -513,7 +594,7 @@ class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
         :returns: Message that was sent to the specified input.
         :rtype: :class:`azure.iot.device.Message`
         """
-        self._validate_receive_api_invoke()
+        self._check_receive_mode_is_api()
 
         if not self._mqtt_pipeline.feature_enabled[constant.INPUT_MSG]:
             await self._enable_feature(constant.INPUT_MSG)
@@ -527,13 +608,18 @@ class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
     async def invoke_method(self, method_params, device_id, module_id=None):
         """Invoke a method from your client onto a device or module client, and receive the response to the method call.
 
-        :param dict method_params: Should contain a method_name, payload, connect_timeout_in_seconds, response_timeout_in_seconds.
+        :param dict method_params: Should contain a methodName (str), payload (str),
+            connectTimeoutInSeconds (int), responseTimeoutInSeconds (int).
         :param str device_id: Device ID of the target device where the method will be invoked.
         :param str module_id: Module ID of the target module where the method will be invoked. (Optional)
 
         :returns: method_result should contain a status, and a payload
         :rtype: dict
         """
+        logger.info(
+            "Invoking {} method on {}{}".format(method_params["methodName"], device_id, module_id)
+        )
+
         invoke_method_async = async_adapter.emulate_async(self._http_pipeline.invoke_method)
         callback = async_adapter.AwaitableCallback(return_arg_name="invoke_method_response")
         await invoke_method_async(device_id, method_params, callback=callback, module_id=module_id)
@@ -541,3 +627,15 @@ class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
         method_response = await handle_result(callback)
         logger.info("Successfully invoked method")
         return method_response
+
+    @property
+    def on_message_received(self):
+        """The handler function or coroutine that will be called when an input message is received.
+
+        The function definitionor coroutine should take one positional argument (the
+        :class:`azure.iot.device.Message` object)"""
+        return self._handler_manager.on_message_received
+
+    @on_message_received.setter
+    def on_message_received(self, value):
+        self._generic_handler_setter("on_message_received", constant.INPUT_MSG, value)
