@@ -4,9 +4,17 @@
 # license information.
 # --------------------------------------------------------------------------
 """This module contains an Inbox class for use with an asynchronous client"""
-
+import asyncio
+import threading
 import janus
 from azure.iot.device.iothub.sync_inbox import AbstractInbox
+from . import loop_management
+
+# IMPLEMENTATION NOTE: The janus Queue exists entirely on the "client internal loop",
+# which runs on its own thread. Think of it kind of as a worker loop where async inbox access
+# operations are scheduled, with the results returned back to whatever thread/loop scheduled them.
+# We do this so that it is safe to use inboxes across different threads, in different places.
+# (e.g. customer thread, handler manager thread, callback thread, etc.)
 
 
 class AsyncClientInbox(AbstractInbox):
@@ -17,7 +25,18 @@ class AsyncClientInbox(AbstractInbox):
 
     def __init__(self):
         """Initializer for AsyncClientInbox."""
-        self._queue = janus.Queue()
+
+        # The queue must be instantiated on the client internal loop, but there's no way to do
+        # that at instantiation from a different loop, so instead we make coroutine to do the
+        # task and run it on the client internal loop.
+        # It's not pretty, but it works (newer versions of janus have a loop parameter, but
+        # not the version we are currently locked at)
+        async def make_queue():
+            return janus.Queue()
+
+        loop = loop_management.get_client_internal_loop()
+        fut = asyncio.run_coroutine_threadsafe(make_queue(), loop)
+        self._queue = fut.result()
 
     def __contains__(self, item):
         """Return True if item is in Inbox, False otherwise"""
@@ -44,11 +63,16 @@ class AsyncClientInbox(AbstractInbox):
 
         :returns: An item from the Inbox.
         """
-        return await self._queue.async_q.get()
+        loop = loop_management.get_client_internal_loop()
+        fut = asyncio.run_coroutine_threadsafe(self._queue.async_q.get(), loop)
+        return await asyncio.wrap_future(fut)
 
     def empty(self):
         """Returns True if the inbox is empty, False otherwise
 
+        Note that there is a race condition here, and this may not be accurate. This is because
+        the .empty() operation on a janus queue is not threadsafe.
+
         :returns: Boolean indicating if the inbox is empty
         """
         return self._queue.async_q.empty()
