@@ -8,6 +8,7 @@ Azure IoTHub Device SDK for Python.
 """
 
 import logging
+import deprecation
 from .abstract_clients import (
     AbstractIoTHubClient,
     AbstractIoTHubDeviceClient,
@@ -16,6 +17,7 @@ from .abstract_clients import (
 from .models import Message
 from .inbox_manager import InboxManager
 from .sync_inbox import SyncClientInbox, InboxEmpty
+from . import sync_handler_manager
 from .pipeline import constant as pipeline_constant
 from .pipeline import exceptions as pipeline_exceptions
 from azure.iot.device import exceptions
@@ -72,6 +74,9 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         # **kwargs.
         super(GenericIoTHubClient, self).__init__(**kwargs)
         self._inbox_manager = InboxManager(inbox_type=SyncClientInbox)
+        self._handler_manager = sync_handler_manager.SyncHandlerManager(self._inbox_manager)
+
+        # Set pipeline handlers
         self._mqtt_pipeline.on_connected = CallableWeakMethod(self, "_on_connected")
         self._mqtt_pipeline.on_disconnected = CallableWeakMethod(self, "_on_disconnected")
         self._mqtt_pipeline.on_method_request_received = CallableWeakMethod(
@@ -81,15 +86,46 @@ class GenericIoTHubClient(AbstractIoTHubClient):
             self._inbox_manager, "route_twin_patch"
         )
 
-    def _on_connected(self):
-        """Helper handler that is called upon an iothub pipeline connect"""
-        logger.info("Connection State - Connected")
+    def _enable_feature(self, feature_name):
+        """Enable an Azure IoT Hub feature.
 
-    def _on_disconnected(self):
-        """Helper handler that is called upon an iothub pipeline disconnect"""
-        logger.info("Connection State - Disconnected")
-        self._inbox_manager.clear_all_method_requests()
-        logger.info("Cleared all pending method requests due to disconnect")
+        This is a synchronous call, meaning that this function will not return until the feature
+        has been enabled.
+
+        :param feature_name: The name of the feature to enable.
+            See azure.iot.device.common.pipeline.constant for possible values
+        """
+        logger.info("Enabling feature:" + feature_name + "...")
+        if not self._mqtt_pipeline.feature_enabled[feature_name]:
+            callback = EventedCallback()
+            self._mqtt_pipeline.enable_feature(feature_name, callback=callback)
+            callback.wait_for_completion()
+
+            logger.info("Successfully enabled feature:" + feature_name)
+        else:
+            # This branch shouldn't be reached, but in case it is, log it
+            logger.info("Feature ({}) already disabled - skipping".format(feature_name))
+
+    def _disable_feature(self, feature_name):
+        """Disable an Azure IoT Hub feature
+
+        This is a synchronous call, meaning that this function will not return until the feature
+        has been disabled.
+
+        :param feature_name: The name of the feature to disable.
+            See azure.iot.device.common.pipeline.constant for possible values
+        """
+        logger.info("Disabling feature: {}...".format(feature_name))
+        if self._mqtt_pipeline.feature_enabled[feature_name]:
+            # Disable the feature if not already disabled
+            callback = EventedCallback()
+            self._mqtt_pipeline.disable_feature(feature_name, callback=callback)
+            callback.wait_for_completion()
+
+            logger.info("Successfully disabled feature: {}".format(feature_name))
+        else:
+            # This branch shouldn't be reached, but in case it is, log it
+            logger.info("Feature ({}) already disabled - skipping".format(feature_name))
 
     def connect(self):
         """Connects the client to an Azure IoT Hub or Azure IoT Edge Hub instance.
@@ -120,6 +156,9 @@ class GenericIoTHubClient(AbstractIoTHubClient):
     def disconnect(self):
         """Disconnect the client from the Azure IoT Hub or Azure IoT Edge Hub instance.
 
+        It is recommended that you make sure to call this function when you are completely done
+        with the your client instance.
+
         This is a synchronous call, meaning that this function will not return until the connection
         to the service has been completely closed.
 
@@ -128,9 +167,33 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         """
         logger.info("Disconnecting from Hub...")
 
+        logger.debug("Executing initial disconnect")
         callback = EventedCallback()
         self._mqtt_pipeline.disconnect(callback=callback)
         handle_result(callback)
+        logger.debug("Successfully executed initial disconnect")
+
+        # Note that in the process of stopping the handlers and resolving pending calls
+        # a user-supplied handler may cause a reconnection to occur
+        logger.debug("Stopping handlers...")
+        self._handler_manager.stop()
+        logger.debug("Successfully stopped handlers")
+
+        # Disconnect again to ensure disconnection has ocurred due to the issue mentioned above
+        logger.debug("Executing secondary disconnect...")
+        callback = EventedCallback()
+        self._mqtt_pipeline.disconnect(callback=callback)
+        handle_result(callback)
+        logger.debug("Successfully executed secondary disconnect")
+
+        # It's also possible that in the (very short) time between stopping the handlers and
+        # the second disconnect, additional items were received (e.g. C2D Message)
+        # Currently, this isn't really possible to accurately check due to a race condition.
+        # It has always been true of this client, even before handlers.
+        # TODO: fix the race condition
+        # However, even if the race condition is addressed, that will only allow us to log that
+        # messages were lost. To actually fix the problem, IoTHub needs to support MQTT5 so that
+        # we can unsubscribe from receiving data.
 
         logger.info("Successfully disconnected from Hub")
 
@@ -171,6 +234,11 @@ class GenericIoTHubClient(AbstractIoTHubClient):
 
         logger.info("Successfully sent message to Hub")
 
+    @deprecation.deprecated(
+        deprecated_in="2.3.0",
+        current_version=device_constant.VERSION,
+        details="We recommend that you use the .on_method_request_received property to set a handler instead",
+    )
     def receive_method_request(self, method_name=None, block=True, timeout=None):
         """Receive a method request via the Azure IoT Hub or Azure IoT Edge Hub.
 
@@ -183,7 +251,7 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         :returns: MethodRequest object representing the received method request, or None if
             no method request has been received by the end of the blocking period.
         """
-        self._validate_receive_api_invoke()
+        self._check_receive_mode_is_api()
 
         if not self._mqtt_pipeline.feature_enabled[pipeline_constant.METHODS]:
             self._enable_feature(pipeline_constant.METHODS)
@@ -227,23 +295,6 @@ class GenericIoTHubClient(AbstractIoTHubClient):
 
         logger.info("Successfully sent method response to Hub")
 
-    def _enable_feature(self, feature_name):
-        """Enable an Azure IoT Hub feature.
-
-        This is a synchronous call, meaning that this function will not return until the feature
-        has been enabled.
-
-        :param feature_name: The name of the feature to enable.
-            See azure.iot.device.common.pipeline.constant for possible values
-        """
-        logger.info("Enabling feature:" + feature_name + "...")
-
-        callback = EventedCallback()
-        self._mqtt_pipeline.enable_feature(feature_name, callback=callback)
-        callback.wait_for_completion()
-
-        logger.info("Successfully enabled feature:" + feature_name)
-
     def get_twin(self):
         """
         Gets the device or module twin from the Azure IoT Hub or Azure IoT Edge Hub service.
@@ -306,6 +357,11 @@ class GenericIoTHubClient(AbstractIoTHubClient):
 
         logger.info("Successfully patched twin")
 
+    @deprecation.deprecated(
+        deprecated_in="2.3.0",
+        current_version=device_constant.VERSION,
+        details="We recommend that you use the .on_twin_desired_properties_patch_received property to set a handler instead",
+    )
     def receive_twin_desired_properties_patch(self, block=True, timeout=None):
         """
         Receive a desired property patch via the Azure IoT Hub or Azure IoT Edge Hub.
@@ -327,7 +383,7 @@ class GenericIoTHubClient(AbstractIoTHubClient):
             received by the end of the blocking period
         :rtype: dict or None
         """
-        self._validate_receive_api_invoke()
+        self._check_receive_mode_is_api()
 
         if not self._mqtt_pipeline.feature_enabled[pipeline_constant.TWIN_PATCHES]:
             self._enable_feature(pipeline_constant.TWIN_PATCHES)
@@ -341,6 +397,46 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         logger.info("twin patch received")
         return patch
 
+    def _generic_handler_setter(self, handler_name, feature_name, new_handler):
+        self._check_receive_mode_is_handler()
+        # Set the handler on the handler manager
+        setattr(self._handler_manager, handler_name, new_handler)
+
+        # Enable the feature if necessary
+        if new_handler is not None and not self._mqtt_pipeline.feature_enabled[feature_name]:
+            self._enable_feature(feature_name)
+
+        # Disable the feature if necessary
+        elif new_handler is None and self._mqtt_pipeline.feature_enabled[feature_name]:
+            self._disable_feature(feature_name)
+
+    @property
+    def on_twin_desired_properties_patch_received(self):
+        """The handler function that will be called when a twin desired properties patch
+        is received.
+
+        The function definition should take one positional argument (the twin patch in the form
+        of a JSON dictionary object)"""
+        return self._handler_manager.on_twin_desired_properties_patch_received
+
+    @on_twin_desired_properties_patch_received.setter
+    def on_twin_desired_properties_patch_received(self, value):
+        self._generic_handler_setter(
+            "on_twin_desired_properties_patch_received", pipeline_constant.TWIN_PATCHES, value
+        )
+
+    @property
+    def on_method_request_received(self):
+        """The handler function that will be called when a method request is received.
+
+        The function definition should take one positional argument (the
+        :class:`azure.iot.device.MethodRequest` object)"""
+        return self._handler_manager.on_method_request_received
+
+    @on_method_request_received.setter
+    def on_method_request_received(self, value):
+        self._generic_handler_setter("on_method_request_received", pipeline_constant.METHODS, value)
+
 
 class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
     """A synchronous device client that connects to an Azure IoT Hub instance.
@@ -364,6 +460,11 @@ class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
             self._inbox_manager, "route_c2d_message"
         )
 
+    @deprecation.deprecated(
+        deprecated_in="2.3.0",
+        current_version=device_constant.VERSION,
+        details="We recommend that you use the .on_message_received property to set a handler instead",
+    )
     def receive_message(self, block=True, timeout=None):
         """Receive a message that has been sent from the Azure IoT Hub.
 
@@ -374,7 +475,7 @@ class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
             no method request has been received by the end of the blocking period.
         :rtype: :class:`azure.iot.device.Message` or None
         """
-        self._validate_receive_api_invoke()
+        self._check_receive_mode_is_api()
 
         if not self._mqtt_pipeline.feature_enabled[pipeline_constant.C2D_MSG]:
             self._enable_feature(pipeline_constant.C2D_MSG)
@@ -422,6 +523,18 @@ class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
         handle_result(callback)
         logger.info("Successfully notified blob upload status")
 
+    @property
+    def on_message_received(self):
+        """The handler function that will be called when a message is received.
+
+        The function definition should take one positional argument (the
+        :class:`azure.iot.device.Message` object)"""
+        return self._handler_manager.on_message_received
+
+    @on_message_received.setter
+    def on_message_received(self, value):
+        self._generic_handler_setter("on_message_received", pipeline_constant.C2D_MSG, value)
+
 
 class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
     """A synchronous module client that connects to an Azure IoT Hub or Azure IoT Edge instance.
@@ -489,6 +602,11 @@ class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
 
         logger.info("Successfully sent message to output: " + output_name)
 
+    @deprecation.deprecated(
+        deprecated_in="2.3.0",
+        current_version=device_constant.VERSION,
+        details="We recommend that you use the .on_message_received property to set a handler instead",
+    )
     def receive_message_on_input(self, input_name, block=True, timeout=None):
         """Receive an input message that has been sent from another Module to a specific input.
 
@@ -499,7 +617,7 @@ class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
         :returns: Message that was sent to the specified input, or None if
             no method request has been received by the end of the blocking period.
         """
-        self._validate_receive_api_invoke()
+        self._check_receive_mode_is_api()
 
         if not self._mqtt_pipeline.feature_enabled[pipeline_constant.INPUT_MSG]:
             self._enable_feature(pipeline_constant.INPUT_MSG)
@@ -516,13 +634,17 @@ class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
     def invoke_method(self, method_params, device_id, module_id=None):
         """Invoke a method from your client onto a device or module client, and receive the response to the method call.
 
-        :param dict method_params: Should contain a method_name, payload, connect_timeout_in_seconds, response_timeout_in_seconds.
+        :param dict method_params: Should contain a methodName (str), payload (str),
+            connectTimeoutInSeconds (int), responseTimeoutInSeconds (int).
         :param str device_id: Device ID of the target device where the method will be invoked.
         :param str module_id: Module ID of the target module where the method will be invoked. (Optional)
 
         :returns: method_result should contain a status, and a payload
         :rtype: dict
         """
+        logger.info(
+            "Invoking {} method on {}{}".format(method_params["methodName"], device_id, module_id)
+        )
         callback = EventedCallback(return_arg_name="invoke_method_response")
         self._http_pipeline.invoke_method(
             device_id, method_params, callback=callback, module_id=module_id
@@ -530,3 +652,15 @@ class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
         invoke_method_response = handle_result(callback)
         logger.info("Successfully invoked method")
         return invoke_method_response
+
+    @property
+    def on_message_received(self):
+        """The handler function that will be called when an input message is received.
+
+        The function definition should take one positional argument (the
+        :class:`azure.iot.device.Message` object)"""
+        return self._handler_manager.on_message_received
+
+    @on_message_received.setter
+    def on_message_received(self, value):
+        self._generic_handler_setter("on_message_received", pipeline_constant.INPUT_MSG, value)
